// facilities for fancy indexing of matrices

namespace mathlib {

// wrapped around indexing objects to allow subscripting
template<class I>
struct matrix_indexer {
	// the argument is not interesting but makes the constants legal
	// without long initializers
	explicit matrix_indexer(int) {}
	matrix_indexer() {}

	template<class T> matrix_indexer<index_sequence<I, T> >
	operator,(matrix_indexer<T>) const
		{ return matrix_indexer<index_sequence<I, T> >(); }

	template<class T> matrix_indexer<index_stack<I, T> >
	operator|(matrix_indexer<T>) const
		{ return matrix_indexer<index_stack<I, T> >(); }

	template<class T> matrix_indexer<index_sequence<T, I> >
	operator()(const matrix_indexer<T>) const
		{ return matrix_indexer<index_sequence<T, I> >(); }
};

template<int RC> struct mask {
	enum { row = 0x000f << RC*4 };
	enum { col = 0x1111 << RC };
	enum { rows_ge = 0xffff & (0xffff << RC*4) };
	enum { rows_le = 0xffff >> (3-RC)*4 };
	enum { col_ge_t = 0xf & (0xf << RC) };
	enum { col_le_t = 0xf >> (3-RC) };
	enum { cols_ge = col_ge_t | (col_ge_t << 4) | (col_ge_t << 8)
		| (col_ge_t << 12) };
	enum { cols_le = col_le_t | (col_le_t << 4) | (col_le_t << 8)
		| (col_le_t << 12) };
};

template<class I, class T> struct index_info {};

// concatenate indexing transformations
template<class I, class J, class A>
struct index_info<index_sequence<I, J>, A> {
	typedef index_info<I, A> i_only;
	typedef index_info<J, i_only> ij;

	static const int rows = ij::rows;
	static const int columns = ij::columns;
	static const bool error = i_only::error || ij::error;
	static const unsigned element_mask = 0;
	static const bool aliasing = i_only::aliasing || ij::aliasing;
	template<int R, int C> struct index {
		typedef typename ij::index<R, C> j;
		enum { rt = j::r };
		enum { ct = j::c };
		typedef typename i_only::index<rt, ct> i;
		enum { r = i::r };
		enum { c = i::c };
	};
};

// stack submatrices
template<class I, class J, class A>
struct index_info<index_stack<I, J>, A> {
	typedef index_info<I, A> i_only;
	typedef index_info<J, A> j_only;

	static const int rows = i_only::rows + j_only::rows;
	static const int columns = i_only::columns;
	static const bool error = i_only::error || j_only::error
		|| i_only::columns != j_only::columns;
	static const unsigned element_mask = i_only::element_mask
		| j_only::element_mask;
	static const bool aliasing = i_only::aliasing || j_only::aliasing
		|| (i_only::element_mask & j_only::element_mask);
	enum { limit = i_only::rows-1 };
	template<int R, int C> struct index {
		typedef typename i_only::index<R, C> i;
		typedef typename j_only::index<R-j_only::rows, C> j;
		enum { r = R > limit ? j::r : i::r };
		enum { c = R > limit ? j::c : i::c };
	};
};

// rearrange elements
template<int X, int Y, int Z, int W, class A>
struct index_info<index_swizzle<X, Y, Z, W>, A> {
	enum { in_range = X < A::rows && Y < A::rows
		&& Z < A::rows && W < A::rows };
	static const int rows = Z == -98 ? 2 : W == -99 ? 3 : 4;
	static const int columns = 1;
	static const bool error = A::columns != 1 || !in_range;
	static const unsigned element_mask = (1<<X) | (1<<Y)
		| (1<<(Z!=-98 ? Z : X)) | (1<<(W!=-99 ? W : X));
	static const bool aliasing = X == Y || X == Z || X == W
		|| Y == Z || Y == W || Z == W;
	template<int R, int C> struct index {
		enum { r = R == 0 ? X : R == 1 ? Y : R == 2 ? Z : W };
		enum { c = 0 };
	};
};

// transpose matrix
template<class A>
struct index_info<index_transpose, A> {
	static const int rows = A::columns;
	static const int columns = A::rows;
	static const bool error = false;
	static const unsigned element_mask = 0xffff;
	static const bool aliasing = false;
	template<int R, int C> struct index {
		enum { r = C };
		enum { c = R };
	};
};

// extract single element
template<int R, int C, class A>
struct index_info<index_element<R, C>, A> {
	static const int rows = 1;
	static const int columns = 1;
	static const bool error = false;
	static const unsigned element_mask = 1 << R+C*4;
	static const bool aliasing = false;
	template<int, int> struct index {
		enum { r = R };
		enum { c = C };
	};
};

// remove row R and column C
template<int R, int C, class A>
struct index_info<index_cofactor<R, C>, A> {
	static const int rows = A::rows-1;
	static const int columns = A::columns-1;
	static const bool error = rows <= 0 || columns <= 0;
	static const unsigned element_mask = 0xffff & ~mask<R>::row
		& ~mask<C>::col;
	static const bool aliasing = false;
	template<int IR, int IC> struct index {
		enum { r = IR >= R ? IR+1 : IR };
		enum { c = IC >= C ? IC+1 : IC };
	};
};

// select submatrix
template<int R1, int C1, int R2, int C2, bool RR, bool RC, class A>
struct index_info<index_submatrix<R1, C1, R2, C2, RR, RC>, A> {
	enum { max_rows = R2-R1+1 };
	enum { max_cols = C2-C1+1 };
	static const int rows = max_rows > A::rows ? A::rows : max_rows;
	static const int columns = max_cols > A::columns
		? A::columns : max_cols;
	static const bool error = rows <= 0 || columns <= 0
		|| RR && max_rows > A::rows
		|| RC && max_cols > A::columns;
	static const unsigned element_mask = mask<R1>::rows_ge
		& mask<R2>::rows_le & mask<C1>::cols_ge & mask<C2>::cols_le;
	static const bool aliasing = false;
	template<int R, int C> struct index {
		enum { r = R + R1 };
		enum { c = C + C1 };
	};
};

// element selection
const matrix_indexer<index_element<0,0> > x(0);
const matrix_indexer<index_element<1,0> > y(0);
const matrix_indexer<index_element<2,0> > z(0);
const matrix_indexer<index_element<3,0> > w(0);
const matrix_indexer<index_element<0,0> > s(0);
const matrix_indexer<index_element<1,0> > t(0);
const matrix_indexer<index_element<2,0> > u(0);
const matrix_indexer<index_element<3,0> > v(0);
const matrix_indexer<index_element<0,0> > r(0);
const matrix_indexer<index_element<1,0> > g(0);
const matrix_indexer<index_element<2,0> > b(0);
const matrix_indexer<index_element<3,0> > a(0);

// transpose
const matrix_indexer<index_transpose> tr(0);

// cartesian coordinate swizzling
const matrix_indexer<index_swizzle<0,1> > xy(0);
const matrix_indexer<index_swizzle<0,2> > xz(0);
const matrix_indexer<index_swizzle<0,3> > xw(0);
const matrix_indexer<index_swizzle<1,2> > yz(0);
const matrix_indexer<index_swizzle<1,3> > yw(0);
const matrix_indexer<index_swizzle<2,3> > zw(0);
const matrix_indexer<index_swizzle<0,1,2> > xyz(0);
const matrix_indexer<index_swizzle<0,1,3> > xyw(0);
const matrix_indexer<index_swizzle<0,2,3> > xzw(0);
const matrix_indexer<index_swizzle<1,2,3> > yzw(0);

// color swizzling
const matrix_indexer<index_swizzle<0,1> > ra(0);
const matrix_indexer<index_swizzle<0,1,2> > rgb(0);
const matrix_indexer<index_swizzle<2,1,0> > bgr(0);
const matrix_indexer<index_swizzle<2,1,0,3> > bgra(0);
const matrix_indexer<index_swizzle<3,0,1,2> > argb(0);
const matrix_indexer<index_swizzle<0,2,1,0> > abgr(0);

// texture coordinate swizzling
const matrix_indexer<index_swizzle<0,1> > st(0);
const matrix_indexer<index_swizzle<0,2> > sz(0);
const matrix_indexer<index_swizzle<0,3> > sw(0);
const matrix_indexer<index_swizzle<1,2> > tu(0);
const matrix_indexer<index_swizzle<1,3> > tv(0);
const matrix_indexer<index_swizzle<2,3> > uv(0);
const matrix_indexer<index_swizzle<0,1,2> > stu(0);
const matrix_indexer<index_swizzle<0,1,3> > stv(0);
const matrix_indexer<index_swizzle<0,2,3> > suv(0);
const matrix_indexer<index_swizzle<1,2,3> > tuv(0);

// submatrix selection
const matrix_indexer<index_submatrix<0,0,1,1> > m22;
const matrix_indexer<index_submatrix<0,0,1,2> > m23;
const matrix_indexer<index_submatrix<0,0,1,3> > m24;
const matrix_indexer<index_submatrix<0,0,2,1> > m32;
const matrix_indexer<index_submatrix<0,0,2,2> > m33;
const matrix_indexer<index_submatrix<0,0,2,3> > m34;
const matrix_indexer<index_submatrix<0,0,3,1> > m42;
const matrix_indexer<index_submatrix<0,0,3,2> > m43;

// row selection
const matrix_indexer<index_submatrix<0,0,0,3,true,false> > r0;
const matrix_indexer<index_submatrix<1,0,1,3,true,false> > r1;
const matrix_indexer<index_submatrix<2,0,2,3,true,false> > r2;
const matrix_indexer<index_submatrix<3,0,3,3,true,false> > r3;
const matrix_indexer<index_submatrix<0,0,1,3,true,false> > r01;
const matrix_indexer<index_submatrix<0,0,2,3,true,false> > r02;
const matrix_indexer<index_submatrix<1,0,2,3,true,false> > r12;
const matrix_indexer<index_submatrix<1,0,3,3,true,false> > r13;
const matrix_indexer<index_submatrix<2,0,3,3,true,false> > r23;

// column selection
const matrix_indexer<index_submatrix<0,0,3,0,false,true> > c0;
const matrix_indexer<index_submatrix<0,1,3,1,false,true> > c1;
const matrix_indexer<index_submatrix<0,2,3,2,false,true> > c2;
const matrix_indexer<index_submatrix<0,3,3,3,false,true> > c3;
const matrix_indexer<index_submatrix<0,0,3,1,false,true> > c01;
const matrix_indexer<index_submatrix<0,0,3,2,false,true> > c02;
const matrix_indexer<index_submatrix<0,1,3,2,false,true> > c12;
const matrix_indexer<index_submatrix<0,1,3,3,false,true> > c13;
const matrix_indexer<index_submatrix<0,2,3,3,false,true> > c23;

} // namespace mathlib
