// unary operator expression templates

namespace mathlib {

template<class Op, class A> struct matrix_unop {};

template<class A>
struct matrix_unop<unop_negate, A> {
	static const int rows = A::rows;
	static const int columns = A::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error;
	static const bool element_op = A::element_op;
	typedef matrix_unop<unop_negate, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
	template<int R, int C> float element() const
		{ return -a.element<R, C>(); }
};

template<class A>
struct matrix_unop<unop_reciprocal, A> {
	static const int rows = A::rows;
	static const int columns = A::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error;
	static const bool element_op = A::element_op;
	typedef matrix_unop<unop_reciprocal, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
	template<int R, int C> float element() const
		{ return 1 / a.element<R, C>(); }
};

template<class A>
struct matrix_unop<unop_adjoint, A> {
	static const int rows = A::columns;
	static const int columns = A::rows;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error;
	static const bool element_op = false;
	typedef matrix_unop<unop_adjoint, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_inverse, A> {
	static const int rows = A::rows;
	static const int columns = A::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || rows != columns;
	static const bool element_op = false;
	typedef matrix_unop<unop_inverse, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_wdiv, A> {
	static const int rows = 3;
	static const int columns = 1;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || A::rows != 4
		|| A::columns != 1;
	static const bool element_op = A::is_storage;
	typedef matrix_unop<unop_wdiv, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
	template<int R, int C> float element() const
		{ return a.element<R, 0>() / a.element<3, 0>(); }
};

template<class A>
struct matrix_unop<unop_normalize, A> {
	static const int rows = A::rows;
	static const int columns = 1;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || A::columns != 1;
	static const bool element_op = false;
	typedef matrix_unop<unop_normalize, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_translation, A> {
	static const int rows = 4;
	static const int columns = 4;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || A::rows != 3 || A::columns != 1;
	static const bool element_op = false;
	typedef matrix_unop<unop_translation, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_scaling, A> {
	static const int rows = 4;
	static const int columns = 4;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error;
	static const bool element_op = false;
	typedef matrix_unop<unop_scaling, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class I, class A>
struct matrix_unop<unop_index<I>, A> {
	typedef index_info<I, A> info;
	static const int rows = info::rows;
	static const int columns = info::columns;
	static const bool is_storage = A::is_storage;
	static const bool is_lvalue = is_storage && !info::aliasing;
	static const bool error = A::error || info::error;
	static const bool element_op = A::element_op && !info::aliasing
		|| is_storage;

	typedef matrix_unop<unop_index<I>, A> base_type;

	typename matrix_ref<A>::noconst_type a;
	matrix_unop(A &a_) : a(a_) {}

	matrix_unop &operator=(const matrix_unop &m) {
		typedef typename assign_dispatch<matrix_unop, 
			matrix_unop>::handler handler;
		handler::go(*this, m);
		return *this;
	}

	template<int R, int C> float element() const {
		typedef typename info::index<R, C> i;
		return a.template element<i::r, i::c>();
	}
	template<int R, int C> float &element() {
		typedef typename info::index<R, C> i;
		return a.template element<i::r, i::c>();
	}
};

template<class I, class A>
struct matrix_unop<unop_cindex<I>, A> {

	typedef index_info<I, A> info;
	static const int rows = info::rows;
	static const int columns = info::columns;
	static const bool is_storage = A::is_storage;
	static const bool is_lvalue = false;
	static const bool error = A::error || info::error;
	static const bool element_op = A::element_op && !info::aliasing
		|| is_storage;

	typedef matrix_unop<unop_cindex<I>, A> base_type;

	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}

	template<int R, int C> float element() const {
		typedef typename info::index<R, C> i;
		return a.template element<i::r, i::c>();
	}
};

} // namespace mathlib
