// specializisations for SSE instructions using GCC intrinsics

namespace mathlib {

// use these to overload at the correct level
typedef tmatrix_dim<3, 1, basic_matrix<3, 1> > v3_dbm;
typedef tmatrix_dim<4, 1, basic_matrix<4, 1> > v4_dbm;
typedef tmatrix_dim<4, 4, basic_matrix<4, 4> > m4_dbm;

// cast to v4sf
inline v4sf &sse(basic_matrix<4, 1> &a)
	{ return reinterpret_cast<v4sf &>(a); }
inline v4sf sse(const basic_matrix<4, 1> &a)
	{ return reinterpret_cast<const v4sf &>(a); }
inline v4sf &sse(matrix_storage<4, 1> &a)
	{ return reinterpret_cast<v4sf &>(a); }
inline v4sf sse(const matrix_storage<4, 1> &a)
	{ return reinterpret_cast<const v4sf &>(a); }
inline v4sf &sse(basic_matrix<3, 1> &a)
	{ return reinterpret_cast<v4sf &>(a); }
inline v4sf sse(const basic_matrix<3, 1> &a)
	{ return reinterpret_cast<const v4sf &>(a); }
inline v4sf &sse(matrix_storage<3, 1> &a)
	{ return reinterpret_cast<v4sf &>(a); }
inline v4sf sse(const matrix_storage<3, 1> &a)
	{ return reinterpret_cast<const v4sf &>(a); }

// cast to v4sf[]
inline v4sf *ssev(basic_matrix<4, 4> &a)
	{ return reinterpret_cast<v4sf *>(&a); }
inline const v4sf *ssev(const basic_matrix<4, 4> &a)
	{ return reinterpret_cast<const v4sf *>(&a); }

// fill v4sf with single value
inline v4sf ssefill(float f) {
	v4sf t = __builtin_ia32_loadss(&f);
	return __builtin_ia32_shufps(t, t, 0x00);
}
// extract first value
inline float float0(v4sf t) {
	float f;
	__builtin_ia32_storess(&f, t);
	return f;
}

// add first 3 elements to first
inline v4sf sum3(v4sf v) {
	v4sf ty = __builtin_ia32_shufps(v, v, 0x55);
	v4sf tz = __builtin_ia32_movhlps(v, v);
	v = __builtin_ia32_addss(v, ty);
	return __builtin_ia32_addss(v, tz);
}
// add all elements to first
inline v4sf sum4(v4sf v) {
	v4sf t = __builtin_ia32_movhlps(v, v);
	v = __builtin_ia32_addps(v, t);
	t = __builtin_ia32_shufps(v, v, 0x55);
	return __builtin_ia32_addss(v, t);
}

// cast to vector<>
inline vector<3> vec3(v4sf v)
	{ return reinterpret_cast<const vector<3> &>(v); }
inline vector<4> vec4(v4sf v)
	{ return reinterpret_cast<const vector<4> &>(v); }

inline matrix_storage<4, 1>::matrix_storage(const matrix_storage &a)
	{ sse(*this) = sse(a); }
inline matrix_storage<4, 1> &
matrix_storage<4, 1>::operator=(const matrix_storage &a) { 
	sse(*this) = sse(a);
	return *this;
}

inline matrix_storage<3, 1>::matrix_storage(const matrix_storage &a)
	{ sse(*this) = sse(a); }
inline matrix_storage<3, 1> &
matrix_storage<3, 1>::operator=(const matrix_storage &a) { 
	sse(*this) = sse(a);
	return *this;
}

inline const vector<4>
operator*(const m4_dbm &a, const v4_dbm &b) {
	v4sf v, f1, f2, f3, f4;
	v = sse(b);
	f1 = __builtin_ia32_shufps(v, v, 0x00);
	f2 = __builtin_ia32_shufps(v, v, 0x55);
	f3 = __builtin_ia32_shufps(v, v, 0xaa);
	f4 = __builtin_ia32_shufps(v, v, 0xff);
	v4sf c1, c2, c3, c4;
	c1 = __builtin_ia32_mulps(f1, ssev(a)[0]);
	c2 = __builtin_ia32_mulps(f2, ssev(a)[1]);
	c3 = __builtin_ia32_mulps(f3, ssev(a)[2]);
	c4 = __builtin_ia32_mulps(f4, ssev(a)[3]);
	v4sf s1, s2;
	s1 = __builtin_ia32_addps(c1, c2);
	s2 = __builtin_ia32_addps(c3, c4);
	return vec4(__builtin_ia32_addps(s1, s2));
}

inline const vector<4>
reciprocal(const v4_dbm &a) {
	return vec4(__builtin_ia32_rcpps(sse(a)));
}
inline const vector<3>
reciprocal(const v3_dbm &a) {
	return vec3(__builtin_ia32_rcpps(sse(a)));
}

inline float dot(const v4_dbm &a, const v4_dbm &b) {
	v4sf t = __builtin_ia32_mulps(sse(a), sse(b));
	return float0(sum4(t));
}
inline float dot(const v3_dbm &a, const v3_dbm &b) {
	v4sf t = __builtin_ia32_mulps(sse(a), sse(b));
	return float0(sum3(t));
}

inline vector<3> wdiv(const v4_dbm &a) {
	v4sf in = sse(a);
	v4sf f = __builtin_ia32_shufps(in, in, 0xff);
	return vec3(__builtin_ia32_divps(in, f));
}

inline vector<4> norm(const v4_dbm &a) {
	v4sf in = sse(a);
	v4sf t = __builtin_ia32_mulps(in, in);
	t = sum4(t);
	t = __builtin_ia32_sqrtss(t);
	t = __builtin_ia32_shufps(t, t, 0x00);
	return vec4(__builtin_ia32_divps(in, t));
}
inline vector<3> norm(const v3_dbm &a) {
	v4sf in = sse(a);
	v4sf t = __builtin_ia32_mulps(in, in);
	t = sum3(t);
	t = __builtin_ia32_sqrtss(t);
	t = __builtin_ia32_shufps(t, t, 0x00);
	return vec3(__builtin_ia32_divps(in, t));
}

inline float abs_squared(const v4_dbm &a) {
	v4sf t = __builtin_ia32_mulps(sse(a), sse(a));
	return float0(sum4(t));
}
inline float abs_squared(const v3_dbm &a) {
	v4sf t = __builtin_ia32_mulps(sse(a), sse(a));
	return float0(sum3(t));
}

inline float sum(const v4_dbm &a)
	{ return float0(sum4(sse(a))); }
inline float sum(const v3_dbm &a)
	{ return float0(sum3(sse(a))); }

inline const vector<4>
operator+(const v4_dbm &a, const v4_dbm &b)
	{ return vec4(__builtin_ia32_addps(sse(a), sse(b))); }
inline const vector<3>
operator+(const v3_dbm &a, const v3_dbm &b)
	{ return vec3(__builtin_ia32_addps(sse(a), sse(b))); }

inline const vector<4>
operator-(const v4_dbm &a, const v4_dbm &b)
	{ return vec4(__builtin_ia32_subps(sse(a), sse(b))); }
inline const vector<3>
operator-(const v3_dbm &a, const v3_dbm &b)
	{ return vec3(__builtin_ia32_subps(sse(a), sse(b))); }

inline const vector<4>
operator*(const v4_dbm &a, const v4_dbm &b)
	{ return vec4(__builtin_ia32_mulps(sse(a), sse(b))); }
inline const vector<3>
operator*(const v3_dbm &a, const v3_dbm &b)
	{ return vec3(__builtin_ia32_mulps(sse(a), sse(b))); }

inline const vector<4>
operator*(const v4_dbm &a, float b)
	{ return vec4(__builtin_ia32_mulps(sse(a), ssefill(b))); }
inline const vector<3>
operator*(const v3_dbm &a, float b)
	{ return vec3(__builtin_ia32_mulps(sse(a), ssefill(b))); }

inline const vector<4>
operator*(float a, const v4_dbm &b)
	{ return vec4(__builtin_ia32_mulps(ssefill(a), sse(b))); }
inline const vector<3>
operator*(float a, const v3_dbm &b)
	{ return vec3(__builtin_ia32_mulps(ssefill(a), sse(b))); }

inline const vector<4>
operator/(const v4_dbm &a, const v4_dbm &b)
	{ return vec4(__builtin_ia32_divps(sse(a), sse(b))); }
inline const vector<3>
operator/(const v3_dbm &a, const v3_dbm &b)
	{ return vec3(__builtin_ia32_divps(sse(a), sse(b))); }

inline const vector<4>
operator/(const v4_dbm &a, float b)
	{ return vec4(__builtin_ia32_divps(sse(a), ssefill(b))); }
inline const vector<3>
operator/(const v3_dbm &a, float b)
	{ return vec3(__builtin_ia32_divps(sse(a), ssefill(b))); }

// copy-assign
template<>
struct simd_assign_specialized<unop_copy, bm44,
	matrix_binop<binop_mmul<4>, bm44, bm44> > { enum { value = 1 }; };
template<>
struct simd_assign_specialized<unop_copy, bm41,	float_keeper>
	{ enum { value = 1 }; };
template<>
struct simd_assign_specialized<unop_copy, bm31,	float_keeper>
	{ enum { value = 1 }; };
template<>
struct simd_assign_specialized<unop_copy, bm44,	zero_assigner_type>
	{ enum { value = 1 }; };
template<>
struct simd_assign_specialized<unop_copy, bm41,	zero_assigner_type>
	{ enum { value = 1 }; };
template<>
struct simd_assign_specialized<unop_copy, bm31,	zero_assigner_type>
	{ enum { value = 1 }; };
template<>
struct simd_assign_handler<unop_copy> {
	
	static void go(bm44 &, matrix_binop<binop_mmul<4>, bm44, bm44>);

	static void go(bm41 &d, float_keeper s) { sse(d) = ssefill(s.f); }
	static void go(bm31 &d, float_keeper s) { sse(d) = ssefill(s.f); }
	static void go(bm31 &d, zero_assigner_type)
		{ asm ( "pxor %0, %0" : "=x" (sse(d)) ); }
	static void go(bm41 &d, zero_assigner_type)
		{ asm ( "pxor %0, %0" : "=x" (sse(d)) ); }
	static void go(bm44 &d, zero_assigner_type) {
		v4sf t;
		asm ( "pxor %0, %0" : "=x" (t) );
		ssev(d)[0] = t;
		ssev(d)[1] = t;
		ssev(d)[2] = t;
		ssev(d)[3] = t;
	}
};

// add-assign
template<> struct simd_assign_specialized<binop_add, bm44, bm44>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_add, bm41, bm41>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_add, bm31, bm31>
	{ enum { value = 1 }; };
template<>
struct simd_assign_handler<binop_add> {

	static void go(bm41 &d, const bm41 & MATHLIB_RESTRICT s)
		{ sse(d) = __builtin_ia32_addps(sse(s), sse(d)); }
	static void go(bm31 &d, const bm31 & MATHLIB_RESTRICT s)
		{ sse(d) = __builtin_ia32_addps(sse(s), sse(d)); }
	static void go(bm44 &d, const bm44 & MATHLIB_RESTRICT s) {
		ssev(d)[0] = __builtin_ia32_addps(ssev(s)[0], ssev(d)[0]);
		ssev(d)[1] = __builtin_ia32_addps(ssev(s)[1], ssev(d)[1]);
		ssev(d)[2] = __builtin_ia32_addps(ssev(s)[2], ssev(d)[2]);
		ssev(d)[3] = __builtin_ia32_addps(ssev(s)[3], ssev(d)[3]);
	}
};

// sub-assign
template<> struct simd_assign_specialized<binop_sub, bm44, bm44>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_sub, bm41, bm41>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_sub, bm31, bm31>
	{ enum { value = 1 }; };
template<>
struct simd_assign_handler<binop_sub> {

	static void go(bm41 &d, const bm41 & MATHLIB_RESTRICT s)
		{ sse(d) = __builtin_ia32_subps(sse(s), sse(d)); }
	static void go(bm31 &d, const bm31 & MATHLIB_RESTRICT s)
		{ sse(d) = __builtin_ia32_subps(sse(s), sse(d)); }
	static void go(bm44 &d, const bm44 & MATHLIB_RESTRICT s) {
		ssev(d)[0] = __builtin_ia32_subps(ssev(s)[0], ssev(d)[0]);
		ssev(d)[1] = __builtin_ia32_subps(ssev(s)[1], ssev(d)[1]);
		ssev(d)[2] = __builtin_ia32_subps(ssev(s)[2], ssev(d)[2]);
		ssev(d)[3] = __builtin_ia32_subps(ssev(s)[3], ssev(d)[3]);
	}
};

// mul-assign
template<> struct simd_assign_specialized<binop_mul, bm44, bm44>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_mul, bm41, bm41>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_mul, bm31, bm31>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_mul, bm41, float_keeper>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_mul, bm31, float_keeper>
	{ enum { value = 1 }; };
template<>
struct simd_assign_handler<binop_mul> {

	static void go(bm41 &d, const bm41 & MATHLIB_RESTRICT s)
		{ sse(d) = __builtin_ia32_mulps(sse(s), sse(d)); }
	static void go(bm31 &d, const bm31 & MATHLIB_RESTRICT s)
		{ sse(d) = __builtin_ia32_mulps(sse(s), sse(d)); }
	static void go(bm41 &d, float_keeper s)
		{ sse(d) = __builtin_ia32_mulps(ssefill(s.f), sse(d)); }
	static void go(bm31 &d, float_keeper s)
		{ sse(d) = __builtin_ia32_mulps(ssefill(s.f), sse(d)); }
	static void go(bm44 &d, const bm44 & MATHLIB_RESTRICT s) {
		ssev(d)[0] = __builtin_ia32_mulps(ssev(s)[0], ssev(d)[0]);
		ssev(d)[1] = __builtin_ia32_mulps(ssev(s)[1], ssev(d)[1]);
		ssev(d)[2] = __builtin_ia32_mulps(ssev(s)[2], ssev(d)[2]);
		ssev(d)[3] = __builtin_ia32_mulps(ssev(s)[3], ssev(d)[3]);
	}
};

// div-assign
template<> struct simd_assign_specialized<binop_div, bm44, bm44>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_div, bm41, bm41>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_div, bm31, bm31>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_div, bm41, float_keeper>
	{ enum { value = 1 }; };
template<> struct simd_assign_specialized<binop_div, bm31, float_keeper>
	{ enum { value = 1 }; };
template<>
struct simd_assign_handler<binop_div> {

	static void go(bm41 &d, const bm41 & MATHLIB_RESTRICT s)
		{ sse(d) = __builtin_ia32_divps(sse(s), sse(d)); }
	static void go(bm31 &d, const bm31 & MATHLIB_RESTRICT s)
		{ sse(d) = __builtin_ia32_divps(sse(s), sse(d)); }
	static void go(bm41 &d, float_keeper s)
		{ sse(d) = __builtin_ia32_divps(ssefill(s.f), sse(d)); }
	static void go(bm31 &d, float_keeper s)
		{ sse(d) = __builtin_ia32_divps(ssefill(s.f), sse(d)); }
	static void go(bm44 &d, const bm44 & MATHLIB_RESTRICT s) {
		ssev(d)[0] = __builtin_ia32_divps(ssev(s)[0], ssev(d)[0]);
		ssev(d)[1] = __builtin_ia32_divps(ssev(s)[1], ssev(d)[1]);
		ssev(d)[2] = __builtin_ia32_divps(ssev(s)[2], ssev(d)[2]);
		ssev(d)[3] = __builtin_ia32_divps(ssev(s)[3], ssev(d)[3]);
	}
};

} // namespace mathlib
